from datetime import timedelta
from urllib.parse import urljoin, urlparse

import arrow
import flask
import flask_login
import sqlalchemy as sa
import string
import typing
import wtforms
from blazeutils import tolist
from keg.db import db
from sqlalchemy_utils import EmailType

from keg_auth import forms
from keg_auth.extensions import flash, lazy_gettext as _
from keg_auth.libs import get_domain_from_email
from keg_auth.model import get_username_key, get_username
from keg_auth.model.entity_registry import RegistryError

try:
    import flask_wtf
except ImportError:
    pass  # pragma: no cover

try:
    import flask_jwt_extended
except ImportError:
    pass  # pragma: no cover

try:
    import ldap
except ImportError:
    pass  # pragma: no cover


class UserNotFound(Exception):
    pass


class UserInactive(Exception):
    def __init__(self, user):
        self.user = user


class UserInvalidAuth(Exception):
    def __init__(self, user):
        self.user = user


class AttemptBlocked(Exception):
    pass


class RequestLoader(object):
    """ Generic loader interface for determining if a user should be logged in"""

    def __init__(self, app):
        self.user_ent = app.auth_manager.entity_registry.user_cls

    @classmethod
    def get_identifier(cls):
        return cls.__name__.lower().replace('requestloader', '')


class LoginAuthenticator(object):
    """ Manages verification of users as well as relevant view-layer logic

        Relevant auth views (login, verification, resets, etc.) get passed through to responders
        on this layer, to process and render for the specific type of authentication happening.

        For example, a password authenticator will want a user/password login form, but other types
        like oauth may get a different form entirely (and handle resets differently, etc.).

        `responder_cls` is a key/value store for associating view keys with responder classes. If a
        view key is not present, we assume that view is not relevant to the authenticator, and the
        view itself will return 404.
    """

    authentication_failure_redirect = True
    responder_cls = {}

    def __init__(self, app):
        self.user_ent = app.auth_manager.entity_registry.user_cls
        self.responders = {}

        self.init_responders()

    def init_responders(self):
        for key, cls in self.responder_cls.items():
            self.responders[key] = cls(self)

    def get_responder(self, key):
        return self.responders.get(key)

    def is_domain_excluded(self, login_id):
        return False


class ViewResponder(object):
    """ View-layer logic wrapper for use in the Authenticator

        Responder should be combined with needed mixins for various functionality (forms,
        logins, etc.).

        Expected to have methods named for the request method (get, post, etc.)

        `template_name` is passed to `flask.render_template` by default
    """
    template_name = None
    _csrf_custom_handling = False

    def __init__(self, parent):
        self.template_args = {}
        self.parent = parent

    def assign(self, key, value):
        self.template_args[key] = value

    def render(self):
        return flask.render_template(self.template_name, **self.template_args)

    def handle_csrf(self):
        """ For some views that are rate-limited, we want to log all attempts, including
        those that would fail CSRF validation. Because of this, we need to circumvent
        flask-csrf's default before-request hook. The auth manager will work to exempt
        any of our auth endpoints whose class is marked with _csrf_custom_handling.

        If CSRF fails, ensure the attempt is logged, and then raise the error.

        If CSRF succeeds, the ensuing view responder methods should do any appropriate
        logging.
        """
        app = flask.current_app

        if (
            'csrf' not in app.extensions
            or not app.config['WTF_CSRF_ENABLED']
            or not app.config['WTF_CSRF_CHECK_DEFAULT']
            or flask.request.method not in app.config['WTF_CSRF_METHODS']
        ):
            return

        try:
            app.extensions['csrf'].protect()
        except flask_wtf.csrf.CSRFError:
            self.on_csrf_error()
            raise

    def on_csrf_error(self):
        pass

    def __call__(self, *args, **kwargs):
        if self._csrf_custom_handling:
            self.handle_csrf()

        method_name = flask.request.method.lower()
        method_obj = getattr(self, method_name, None)
        if not method_obj:
            raise NotImplementedError(method_name)
        resp = method_obj(*args, **kwargs)

        return resp or self.render()

    def head(self, *args, **kwargs):
        valid_methods = []
        for method in ('get', 'post'):
            if hasattr(self, method):
                valid_methods.append(method.upper())
        return flask.abort(405, valid_methods=valid_methods)


class UserResponderMixin(object):
    flash_invalid_user = _('No user account matches: {}'), 'error'
    flash_unverified_user = _(
        'The user account "{}" has an unverified email address.  Please check'
        ' your email for a verification link from this website.  Or, use the "forgot'
        ' password" link to verify the account.'
    ), 'error'
    flash_disabled_user = _(
        'The user account "{}" has been disabled.  Please contact this'
        ' site\'s administrators for more information.'
    ), 'error'

    def on_inactive_user(self, user):
        if flask.current_app.auth_manager.mail_manager and not user.is_verified:
            if self.flash_unverified_user:
                message, category = self.flash_unverified_user
                flash(message.format(user.email), category)
                return  # preventing unverified users to also get the disabled alert

        if not user.is_active:
            self.on_disabled_user(user)

    def on_invalid_user(self, username):
        if self.flash_invalid_user:
            message, category = self.flash_invalid_user
            flash(message.format(username), category)

    def on_disabled_user(self, user):
        if self.flash_disabled_user:
            message, category = self.flash_disabled_user
            flash(message.format(user.display_value), category)


class LoginResponderMixin(UserResponderMixin):
    """ Wrap user authentication view-layer logic

        Flash messages, what to do when a user has been authenticated (by whatever method the
        parent authenticator uses), redirects to a safe URL after login, etc.
    """
    url = '/login'
    flash_success = _('Login successful.'), 'success'

    @staticmethod
    def is_safe_url(target):
        """Returns `True` if the target is a valid URL for redirect"""
        # from http://flask.pocoo.org/snippets/62/
        ref_url = urlparse(flask.request.host_url)
        test_url = urlparse(urljoin(flask.request.host_url, target))
        return (
            test_url.scheme in ('http', 'https')
            and ref_url.netloc == test_url.netloc
        )

    def on_success(self, user):
        flask_login.login_user(user)
        if self.flash_success:
            flash(*self.flash_success)

        # support Flask-Login "next" parameter
        next_parameter = flask.request.values.get('next')
        if flask.current_app.config.get('USE_SESSION_FOR_NEXT'):
            next_parameter = flask.session.get('next')
        if next_parameter and self.is_safe_url(next_parameter):
            redirect_to = next_parameter
        else:
            redirect_to = flask.current_app.auth_manager.url_for('after-login')

        return flask.redirect(redirect_to)


class FormResponderMixin(object):
    """ Wrap form usage for auth responders, contains GET and POST handlers"""
    flash_form_error = _('The form has errors, please see below.'), 'error'
    form_cls = None
    page_title = None

    def on_form_error(self, form):
        if self.flash_form_error:
            flash(*self.flash_form_error)

    def on_form_valid(self, form):
        raise NotImplementedError  # pragma: no cover

    def create_form(self):
        return self.form_cls()

    def assign_template_vars(self, form):
        title_var = flask.current_app.config.get('KEGAUTH_TEMPLATE_TITLE_VAR')
        self.assign('form', form)
        self.assign('form_action_text', self.page_title)
        self.assign(title_var, self.page_title)
        self.assign('page_heading', self.page_title)

    def get(self, *args, **kwargs):
        form = self.create_form()
        self.assign_template_vars(form)

    def post(self, *args, **kwargs):
        form = self.create_form()
        if form.validate():
            resp = self.on_form_valid(form)
            if resp is not None:
                return resp
        else:
            self.on_form_error(form)

        self.assign_template_vars(form)


class AttemptLimitMixin(object):
    @property
    def attempt_ent(self):
        return flask.current_app.auth_manager.entity_registry.attempt_cls

    def should_limit_attempts(self):
        if not flask.current_app.config.get('KEGAUTH_ATTEMPT_LIMIT_ENABLED'):
            # limiting can be turned off by config
            return False
        try:
            # verify the attempt entity has been configured for logging attempts
            self.attempt_ent
        except RegistryError:
            raise Exception('Rate limiting is enabled, but the attempt entity is not registered')
        return True

    def check_blocking(self, user_input, success=False):
        """Generic blocking method that will create an attempt log, and notify the calling
        method (via exception) if the attempt is blocked.

        Returns the attempt log if not blocked. If ``success`` is left False (default), The
        calling method will be responsible to set the attempt log's success flag as needed.
        """
        attempt_log = None
        if self.should_limit_attempts():
            if self.is_attempt_blocked(user_input):
                # If we are rate-limiting this attempt, we don't want to proceed with validation.
                # Validating may still allow brute forcing by measuring response time.
                attempt_log = self.log_attempt(user_input, success=False, is_during_lockout=True)
                self.on_attempt_blocked()
                raise AttemptBlocked
            else:
                attempt_log = self.log_attempt(user_input, success=success)
        return attempt_log

    @property
    def should_filter_ip(self):
        return flask.current_app.config.get('KEGAUTH_ATTEMPT_IP_LIMIT', False)

    def get_input_filters(self, username):
        input_filters = self.attempt_ent.user_input == username
        if self.should_filter_ip and self.get_request_remote_addr():
            input_filters = sa.sql.or_(
                input_filters,
                self.attempt_ent.source_ip == self.get_request_remote_addr()
            )
        return input_filters

    def get_last_limiting_attempt(self, username):
        '''
        Get the last attempt that counts toward the limit count. Attempts that count
        toward the limit before this attempt will be counted to determine if this
        attempt caused a lockout.

        For login, this will be the last failed attempt.
        For password reset, this will be the last attempt.
        '''
        raise NotImplementedError()  # pragma: no cover

    def get_limiting_attempt_count(self, before_time, username):
        '''
        Return the number of attempts that count toward the limit up to before_time.
        '''
        raise NotImplementedError  # pragma: no cover

    def is_attempt_blocked(self, username):
        last_limiting_attempt = self.get_last_limiting_attempt(username)

        if last_limiting_attempt:
            limiting_attempt_count = self.get_limiting_attempt_count(
                last_limiting_attempt.datetime_utc,
                username
            )
            # A failed attempt has caused a lockout so we should check if the lockout
            # period has passed.
            if limiting_attempt_count >= self.get_attempt_limit():
                lockout_delta = timedelta(seconds=self.get_attempt_lockout_period())
                return arrow.utcnow() - last_limiting_attempt.datetime_utc <= lockout_delta

        # We are not in a lockout period so block attempts if there are too many
        # limiting attempts in the timespan until now.
        limiting_in_timespan_count = self.get_limiting_attempt_count(arrow.utcnow(), username)
        return limiting_in_timespan_count >= self.get_attempt_limit()

    def log_attempt(self, username, *, success=True, is_during_lockout=False):
        attempt = self.attempt_ent(
            attempt_type=self.get_attempt_type(),
            user_input=username,
            success=success,
            is_during_lockout=is_during_lockout,
            datetime_utc=arrow.utcnow(),
        )
        if flask.has_request_context():
            attempt.source_ip = self.get_request_remote_addr()

        db.session.add(attempt)
        db.session.commit()
        return attempt

    def update_attempt(self, attempt, **kwargs):
        self.attempt_ent.edit(attempt.id, **kwargs)

    @staticmethod
    def get_request_remote_addr():
        return flask.request.remote_addr

    def on_attempt_blocked(self):
        flash_message = self.get_flash_attempts_limit_reached()
        if flash_message:
            flash(*flash_message)

    def get_flash_attempts_limit_reached(self):
        raise NotImplementedError  # pragma: no cover

    def get_attempt_type(self):
        raise NotImplementedError  # pragma: no cover

    def get_config_value(self, key):
        """Fetch the config value for the view specified in get_attempt_type.

        Checks the view-specific key first, and if the value is not configured,
        falls to the default for the key.

        Example: get_attempt_type returns login, and key is attempt_limit.
        - Checks for KEGAUTH_LOGIN_ATTEMPT_LIMIT first
        - If that is not set, falls back to KEGAUTH_ATTEMPT_LIMIT
        """
        view_key = self.get_attempt_type().upper()
        key = key.upper()
        return flask.current_app.config.get(
            f'KEGAUTH_{view_key}_{key}',
            flask.current_app.config.get(f'KEGAUTH_{key}')
        )

    def get_attempt_limit(self):
        return self.get_config_value('attempt_limit')

    def get_attempt_timespan(self):
        return self.get_config_value('attempt_timespan')

    def get_attempt_lockout_period(self):
        return self.get_config_value('attempt_lockout')


class PasswordSetterResponderBase(FormResponderMixin, ViewResponder):
    """ Base logic for resetting passwords and verifying accounts via token"""
    form_cls = forms.SetPassword
    template_name = 'keg-auth/set-password.html'
    flash_invalid_token = _(
        'Authentication token was invalid or expired.  Please fill out the'
        ' form below to get a new token.'
    ), 'error'

    def __call__(self, *args, **kwargs):
        if not flask.current_app.auth_manager.mail_manager:
            flask.abort(404)

        self.user_loader(kwargs.get('user_id'))
        self.token = kwargs.get('token')
        self.pre_method()

        return super(PasswordSetterResponderBase, self).__call__(*args, **kwargs)

    def flash_and_redirect(self, flash_parts, auth_ident):
        if flash_parts:
            flash(*flash_parts)
        redirect_to = flask.current_app.auth_manager.url_for(auth_ident)
        flask.abort(flask.redirect(redirect_to))

    def user_loader(self, user_id):
        user_ent = flask.current_app.auth_manager.entity_registry.user_cls
        self.user = user_ent.query.get(user_id)
        if not self.user:
            flask.abort(404)

    def create_form(self):
        return self.form_cls(user=self.user)

    def pre_method(self):
        if not self.user.token_verify(self.token):
            resp = self.on_invalid_token()
            # In case on_invalid_token() is replaced and it accidently fails to return a value
            # make sure we change that to a generic 400.
            flask.abort(resp or 400)

    def on_form_valid(self, form):
        new_password = form.password.data
        self.user.change_password(self.token, new_password)
        self.flash_and_redirect(self.flash_success, self.on_success_endpoint)

    def on_invalid_token(self):
        self.flash_and_redirect(self.flash_invalid_token, 'forgot-password')

    def assign_template_vars(self, form):
        super(PasswordSetterResponderBase, self).assign_template_vars(form)
        self.assign('submit_button_text', self.submit_button_text)


class ResetPasswordViewResponder(AttemptLimitMixin, PasswordSetterResponderBase):
    """ Responder for resetting passwords via token on keg-auth logins"""
    url = '/reset-password/<int:user_id>/<token>'
    page_title = _('Complete Password Reset')
    submit_button_text = _('Change Password')
    flash_success = _('Password changed.  Please use the new password to login below.'), 'success'
    on_success_endpoint = 'after-reset'

    def on_form_valid(self, form):
        try:
            self.check_blocking(get_username(self.user), success=True)
        except AttemptBlocked:
            # If we are rate-limiting this attempt, we don't want to proceed with validation.
            # Validating may still allow brute forcing by measuring response time.
            return

        new_password = form.password.data
        self.user.change_password(self.token, new_password)
        self.flash_and_redirect(self.flash_success, self.on_success_endpoint)

    def get_flash_attempts_limit_reached(self):
        return _('Too many password reset attempts.'), 'error'

    def get_attempt_type(self):
        return 'reset'

    def get_last_limiting_attempt(self, username):
        return self.attempt_ent.query.filter_by(
            is_during_lockout=False,
            attempt_type=self.get_attempt_type(),
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

    def get_limiting_attempt_count(self, before_time, username):
        timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())
        return self.attempt_ent.query.filter(
            self.get_input_filters(username),
            self.attempt_ent.is_during_lockout == sa.false(),
            self.attempt_ent.datetime_utc > timespan_start,
            self.attempt_ent.datetime_utc <= before_time,
            self.attempt_ent.attempt_type == self.get_attempt_type(),
        ).count()


class VerifyAccountViewResponder(PasswordSetterResponderBase):
    """ Responder for verifying users via email token for keg-auth logins"""
    url = '/verify-account/<int:user_id>/<token>'
    page_title = _('Verify Account & Set Password')
    submit_button_text = _('Verify & Set Password')
    flash_success = _('Account verified & password set.  Please use the new password to login'
                      ' below.'), 'success'
    on_success_endpoint = 'after-verify-account'


class PasswordFormViewResponder(AttemptLimitMixin, LoginResponderMixin,
                                FormResponderMixin, ViewResponder):
    """ Master responder for username/password-style logins, using a login form"""
    template_name = 'keg-auth/login.html'
    page_title = _('Log In')
    flash_invalid_password = _('Invalid password.'), 'error'
    _csrf_custom_handling = True

    @property
    def form_cls(self):
        return forms.login_form()

    def on_form_error(self, form):
        super().on_form_error(form)
        username = form.login_id.data
        try:
            # ensure we log this even though the form didn't validate
            self.check_blocking(username)
        except AttemptBlocked:
            pass

    def on_csrf_error(self):
        username = flask.request.form.get('login_id', '')
        try:
            # ensure we log this even though the form didn't validate
            self.check_blocking(username)
        except AttemptBlocked:
            pass

    def on_form_valid(self, form):
        username = form.login_id.data
        try:
            attempt = self.check_blocking(username)
        except AttemptBlocked:
            # If we are rate-limiting this attempt, we don't want to proceed with validation.
            # Validating may still allow brute forcing by measuring response time. Skipping it
            # may help mitigate DoS attacks on the login page as password hashing is typically
            # an expensive operation
            return

        try:
            # We want to know if the login attempt was successful so we'll try
            # to verify the user. If the user is verified but the attempt is blocked,
            # mark the attempt as successful and abort.
            user = self.parent.verify_user(
                login_id=form.login_id.data,
                password=form.password.data
            )

            if attempt:
                self.update_attempt(attempt, success=True)

            # User is active and password is verified
            return self.on_success(user)
        except UserNotFound:
            self.on_invalid_user(form.login_id.data)
        except UserInactive as exc:
            self.on_inactive_user(exc.user)
        except UserInvalidAuth as exc:
            self.on_invalid_password(exc.user)

    def on_invalid_password(self, user):
        if self.flash_invalid_password:
            flash(*self.flash_invalid_password)

    def get_flash_attempts_limit_reached(self):
        return _('Too many failed login attempts.'), 'error'

    def get_attempt_type(self):
        return 'login'

    def get_last_limiting_attempt(self, username):
        return self.attempt_ent.query.filter_by(
            success=False,
            is_during_lockout=False,
            attempt_type=self.get_attempt_type(),
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

    def get_limiting_attempt_count(self, before_time, username):
        last_successful_attempt = self.attempt_ent.query.filter_by(
            success=True,
            is_during_lockout=False,
            attempt_type=self.get_attempt_type()
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

        def is_within_timespan(attempt):
            timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())
            return attempt.datetime_utc > timespan_start

        if last_successful_attempt and is_within_timespan(last_successful_attempt):
            timespan_start = last_successful_attempt.datetime_utc
        else:
            timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())

        return self.attempt_ent.query.filter(
            self.get_input_filters(username),
            self.attempt_ent.success == sa.false(),
            self.attempt_ent.is_during_lockout == sa.false(),
            self.attempt_ent.datetime_utc > timespan_start,
            self.attempt_ent.datetime_utc <= before_time,
            self.attempt_ent.attempt_type == self.get_attempt_type(),
        ).count()


class OAuthLoginViewResponder(ViewResponder):
    """ OAuth logins, using a provider via authlib"""
    url = '/login/<name>'

    def get(self, *args, **kwargs):
        name = kwargs.get('name')
        oauth = flask.current_app.auth_manager.oauth
        client = oauth.create_client(name)
        if not client:
            return flask.abort(404)

        redirect_uri = flask.current_app.auth_manager.url_for(
            'oauth-authorize', name=name, _external=True
        )
        return client.authorize_redirect(redirect_uri)

    def head(self, *args, **kwargs):
        return flask.abort(405, valid_methods=['GET'])

    def post(self, *args, **kwargs):
        return flask.abort(405, valid_methods=['GET'])


class OAuthAuthorizeViewResponder(LoginResponderMixin, ViewResponder):
    """ OAuth logins, using a provider via authlib"""
    url = '/oauth-authorize/<name>'
    template_name = 'keg-auth/flash-messages-only.html'

    def get(self, *args, **kwargs):
        name = kwargs.get('name')
        oauth = flask.current_app.auth_manager.oauth
        client = oauth.create_client(name)
        if not client:
            return flask.abort(404)

        oauth_profile = self.parent.select_oauth_profile(name)
        token = client.authorize_access_token()
        userinfo = token.get('userinfo')
        if not userinfo:
            userinfo = client.userinfo()

        login_id = userinfo.get(oauth_profile['id_field'])

        try:
            user = self.parent.verify_user(profile_name=name, login_id=login_id)

            # User is active and password is verified
            return self.on_success(user)
        except UserNotFound:
            self.on_invalid_user(login_id)
        except UserInactive as exc:
            self.on_inactive_user(exc.user)

    def head(self, *args, **kwargs):
        return flask.abort(405, valid_methods=['GET'])


class ForgotPasswordViewResponder(AttemptLimitMixin, UserResponderMixin, FormResponderMixin,
                                  ViewResponder):
    """ Master responder for keg-integrated logins, using an email form"""
    url = '/forgot-password'
    form_cls = forms.ForgotPassword
    page_title = _('Initiate Password Reset')
    template_name = 'keg-auth/forgot-password.html'
    flash_success = _('Please check your email for the link to change your password.'), 'success'
    _csrf_custom_handling = True

    def __call__(self, *args, **kwargs):
        if not flask.current_app.auth_manager.mail_manager:
            flask.abort(404)

        return super(ForgotPasswordViewResponder, self).__call__(*args, **kwargs)

    def on_csrf_error(self):
        try:
            # ensure we log this even though the form didn't validate
            self.check_blocking(flask.request.form.get('email', ''))
        except AttemptBlocked:
            pass

    def on_form_error(self, form):
        super().on_form_error(form)
        try:
            # ensure we log this even though the form didn't validate
            self.check_blocking(form.email.data)
        except AttemptBlocked:
            pass

    def on_form_valid(self, form):
        try:
            attempt = self.check_blocking(form.email.data)
        except AttemptBlocked:
            # If we are rate-limiting this attempt, we don't want to proceed with validation.
            # Validating may still allow brute forcing by measuring response time.
            return

        try:
            user = self.parent.verify_user(login_id=form.email.data, allow_unverified=True)

            if attempt:
                self.update_attempt(attempt, success=True)

            # User is active, take action to initiate password reset
            return self.on_success(user)
        except UserNotFound:
            self.on_invalid_user(form.email.data)
        except UserInactive as exc:
            self.on_disabled_user(exc.user)

    def on_success(self, user):
        self.send_email(user)
        if self.flash_success:
            flash(*self.flash_success)
        redirect_to = flask.current_app.auth_manager.url_for('after-forgot')
        return flask.redirect(redirect_to)

    def send_email(self, user):
        user.token_generate()
        flask.current_app.auth_manager.mail_manager.send_reset_password(user)

    def get_flash_attempts_limit_reached(self):
        return _('Too many failed attempts.'), 'error'

    def get_attempt_type(self):
        return 'forgot'

    def get_last_limiting_attempt(self, username):
        return self.attempt_ent.query.filter_by(
            success=False,
            is_during_lockout=False,
            attempt_type=self.get_attempt_type(),
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

    def get_limiting_attempt_count(self, before_time, username):
        last_successful_attempt = self.attempt_ent.query.filter_by(
            success=True,
            is_during_lockout=False,
            attempt_type=self.get_attempt_type()
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

        def is_within_timespan(attempt):
            timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())
            return attempt.datetime_utc > timespan_start

        if last_successful_attempt and is_within_timespan(last_successful_attempt):
            timespan_start = last_successful_attempt.datetime_utc
        else:
            timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())

        return self.attempt_ent.query.filter(
            self.get_input_filters(username),
            self.attempt_ent.success == sa.false(),
            self.attempt_ent.is_during_lockout == sa.false(),
            self.attempt_ent.datetime_utc > timespan_start,
            self.attempt_ent.datetime_utc <= before_time,
            self.attempt_ent.attempt_type == self.get_attempt_type(),
        ).count()


class LogoutViewResponder(ViewResponder):
    url = '/logout'
    flash_success = _('You have been logged out.'), 'success'

    def get(self):
        flask_login.logout_user()
        if self.flash_success:
            flash(*self.flash_success)
        redirect_to = flask.current_app.auth_manager.url_for('after-logout')
        flask.abort(flask.redirect(redirect_to))


def not_found_view_responder_factory(target_url):
    class NotFoundViewResponder(ViewResponder):
        url = target_url

        def get(self):
            flask.abort(404)

    return NotFoundViewResponder


class RedirectLoginViewResponder(ViewResponder):
    url = '/login'

    def get(self):
        target = flask.current_app.config.get('KEGAUTH_REDIRECT_LOGIN_TARGET')
        if not target or not target.startswith('/'):
            raise Exception('KEGAUTH_REDIRECT_LOGIN_TARGET not set or not relative')
        return flask.redirect(target)


class PasswordAuthenticatorMixin(object):
    """ Username/password authenticators will need a way to verify a user is valid
        prior to making it the current user in flask login """
    responder_cls = {
        'login': PasswordFormViewResponder
    }

    def verify_user(self, login_id=None, password=None):
        raise NotImplementedError

    def verify_password(self, user, password):
        return NotImplementedError


class TokenLoaderMixin(object):
    """ Token authenticators will need a way to generate an access token, which will then be
        loaded in the request to log a user into flask-login """
    authentication_failure_redirect = False

    def create_access_token(self, user):
        raise NotImplementedError


class RedirectAuthenticator(LoginAuthenticator):
    """ Redirects to another source for authentication. Useful for when we have an OAuth source
        in mind for primary auth. We will want to redirect /login there, keep /logout, and
        direct other responder keys to return 404.

        Use KEGAUTH_REDIRECT_LOGIN_TARGET to set the login target."""
    responder_cls = {
        'login': RedirectLoginViewResponder,
        'forgot-password': not_found_view_responder_factory('/forgot-password'),
        'reset-password': not_found_view_responder_factory('/reset-password'),
        'verify-account': not_found_view_responder_factory('/verify-account'),
        'logout': LogoutViewResponder,
    }


class KegAuthenticator(PasswordAuthenticatorMixin, LoginAuthenticator):
    """ Uses username/password authentication with a login form, validates against keg-auth db"""
    responder_cls = {
        'login': PasswordFormViewResponder,
        'forgot-password': ForgotPasswordViewResponder,
        'reset-password': ResetPasswordViewResponder,
        'verify-account': VerifyAccountViewResponder,
        'logout': LogoutViewResponder,
    }

    def __init__(self, app):
        super().__init__(app)

        self.load_domain_exclusions(app)

    def load_domain_exclusions(self, app):
        domain_exclusions = []
        for profile in app.config.get('KEGAUTH_OAUTH_PROFILES', []):
            domain_exclusions.extend(tolist(profile.get('domain_filter', [])))
        self.domain_exclusions = domain_exclusions

    def is_domain_excluded(self, login_id):
        """Domains configured for OAuth access are excluded from the password authenticator.

        Any operations not using ``verify_user`` should check for exclusion."""
        if self.domain_exclusions:
            domain = get_domain_from_email(login_id)
            if domain and domain in self.domain_exclusions:
                return True
        return False

    def verify_user(self, login_id=None, password=None, allow_unverified=False):
        if self.is_domain_excluded(login_id):
            # apply a domain filter before looking up the user record
            raise UserNotFound

        user = self.user_ent.query.filter(
            sa.func.lower(self.user_ent.username) == (login_id.lower() if login_id else None)
        ).one_or_none()

        if not user:
            raise UserNotFound
        if not allow_unverified and not user.is_active:
            raise UserInactive(user)
        if allow_unverified and not user.is_enabled:
            raise UserInactive(user)
        if password and not self.verify_password(user, password):
            raise UserInvalidAuth(user)

        return user

    def verify_password(self, user, password):
        return user.password == password


class OAuthAuthenticator(LoginAuthenticator):
    """ Uses OAuth authentication via authlib, validates user info against keg-auth db"""
    responder_cls = {
        'login': OAuthLoginViewResponder,
        'authorize': OAuthAuthorizeViewResponder,
    }

    def __init__(self, app):
        self.load_profiles(app)
        super().__init__(app)

    def load_profiles(self, app):
        from authlib.integrations.flask_client import OAuth

        app.auth_manager.oauth = oauth = OAuth(app)
        for profile in app.config.get('KEGAUTH_OAUTH_PROFILES'):
            expected_keys = 'domain_filter', 'id_field', 'oauth_client_kwargs'
            if len(set(expected_keys) & set(profile.keys())) != len(expected_keys):
                expected_keys_display = ', '.join(expected_keys)
                raise Exception(
                    f'OAuth profile is missing keys, expects {expected_keys_display}'
                )
            oauth.register(**profile['oauth_client_kwargs'])

    def select_oauth_profile(self, name):
        return next(
            filter(
                lambda profile: profile['oauth_client_kwargs']['name'] == name,
                flask.current_app.config.get('KEGAUTH_OAUTH_PROFILES', [])
            ), None
        )

    def verify_user(self, profile_name=None, login_id=None):
        oauth_profile = self.select_oauth_profile(profile_name)
        if not oauth_profile:
            raise Exception(f'OAuth profile {profile_name} is not configured')

        # Some OAuth providers, like Google, can be configured to only pass through users
        #   from specific domains. But we shouldn't make that assumption for all providers,
        #   and since we have the domain filter also for disallowing logins in KegAuthenticator,
        #   it's best to check it here as well.
        if oauth_profile.get('domain_filter'):
            domain = get_domain_from_email(login_id)
            if domain and domain not in tolist(oauth_profile['domain_filter']):
                raise UserNotFound

        user = self.user_ent.query.filter_by(username=login_id).one_or_none()

        if not user:
            raise UserNotFound

        # Users from domains filtered to OAuth do not follow the usual verification
        #   process. Check for that here - if OAuth source says it's a verified user,
        #   trust it.
        if not user.is_verified:
            user.is_verified = True
            db.session.commit()

        if not user.is_active:
            raise UserInactive(user)

        return user


class LdapAuthenticator(KegAuthenticator):
    """ Uses username/password authentication with a login form, validates against LDAP host

        Most responder types won't be relevant here.
    """
    def verify_user(self, login_id=None, password=None):
        user = self.user_ent.query.filter_by(username=login_id).one_or_none()

        if not user:
            username_key = get_username_key(self.user_ent)
            user = self.user_ent.add(**{username_key: login_id})
        if password and not self.verify_password(user, password):
            raise UserInvalidAuth(user)

        return user

    def verify_password(self, user, password):
        """
        Check the given username/password combination at the application's configured LDAP server.
        Returns `True` if the user authentication is successful, `False` otherwise. NOTE: By
        request, authentication can be bypassed by setting the KEGAUTH_LDAP_TEST_MODE configuration
        setting to `True`. When set, all authentication attempts will succeed!
        """

        if flask.current_app.config.get('KEGAUTH_LDAP_TEST_MODE', False):
            return True

        ldap_url = flask.current_app.config.get('KEGAUTH_LDAP_SERVER_URL')
        if not ldap_url:
            raise Exception(_('No KEGAUTH_LDAP_SERVER_URL configured!'))

        ldap_dn_format = flask.current_app.config.get('KEGAUTH_LDAP_DN_FORMAT')
        if not ldap_dn_format:
            raise Exception(_('No KEGAUTH_LDAP_DN_FORMAT configured!'))

        def ldap_bind(server_url):
            session = ldap.initialize(server_url)

            try:
                dn = ldap_dn_format.format(user.username)
                result = session.simple_bind_s(dn, password)
                del session
                return bool(
                    result
                    and len(result)
                    and result[0] == ldap.RES_BIND
                )
            except (ldap.INVALID_CREDENTIALS, ldap.INVALID_DN_SYNTAX):
                return False

        if isinstance(ldap_url, str):
            return ldap_bind(ldap_url)

        # We have a list of servers.
        for server_url in ldap_url:
            if ldap_bind(server_url):
                return True

        return False


class JwtRequestLoader(TokenLoaderMixin, RequestLoader):
    """ Loader for JWT tokens contained in the Authorization header.

        Requires flask-jwt-extended (`pip install keg-auth[jwt]`)"""
    def __init__(self, app):
        super(JwtRequestLoader, self).__init__(app)

        self.jwt_manager = jwt_manager = flask_jwt_extended.JWTManager()
        jwt_manager.init_app(app)

        @jwt_manager.user_identity_loader
        def user_identity_loader(user):
            """
            Serialize a user entity to the JWT token
            This method is the complement of `user_lookup_loader`
            """
            return user.session_key

        @jwt_manager.user_lookup_loader
        def user_loader_callback_loader(jwt_header, jwt_data):
            """
            Load a user entity from the JWT token
            This method is the complement of `user_identity_loader`

            Note, if user is not found or inactive, fail silently - user just won't get loaded
            """
            data_key = flask.current_app.config.get('JWT_IDENTITY_CLAIM')
            return self.user_ent.get_by(session_key=jwt_data[data_key], is_active=True)

    @staticmethod
    def get_authenticated_user():
        try:
            if flask_jwt_extended.verify_jwt_in_request() is None:
                return None
            user = flask_jwt_extended.get_current_user()
            flask_login.login_user(user)
            return user
        except flask_jwt_extended.exceptions.JWTExtendedException:
            return None

    def create_access_token(self, user):
        return flask_jwt_extended.create_access_token(user)


class TokenRequestLoader(RequestLoader):
    authentication_failure_redirect = False

    def get_authenticated_user(self):
        token = flask.request.headers.get('X-Auth-Token')

        if token is None:
            return

        user = self.user_ent.get_by_token(token)

        if user is None:
            return

        flask_login.login_user(user)
        return user


class PasswordCharset(typing.NamedTuple):
    name: str
    alphabet: str


class PasswordPolicyError(Exception):
    pass


PasswordCheckFunction = typing.Callable[[str, typing.Any], None]


class PasswordPolicy:
    """
    A base class that defines password requirements for the application.
    This class defines some basic, common validations and can be extended or limited by subclassing.

    To define additional password checks, create a method on your subclass that accepts a
    password string and a user entity object and raises PasswordPolicyError if the password does
    not meet the requirement you intend to check. Then override password_checks to add your method
    to the returned list of methods.

    To remove a password check that is enabled by default, override password_checks and return only
    the methods you wish to use.

    Default settings are based on NIST guidelines and some common restrictions.
    """

    """
    Minimum password length for check_length validation
    """
    min_length: int = 8

    """
    Character sets used for checking minimum "complexity" in check_character_set validation
    """
    required_char_types: typing.List[PasswordCharset] = [
        PasswordCharset(_('lowercase letter'), string.ascii_lowercase),
        PasswordCharset(_('uppercase letter'), string.ascii_uppercase),
        PasswordCharset(_('number'), string.digits),
        PasswordCharset(_('symbol'), ''.join(sorted(
            set(string.printable)
            - set(string.whitespace + string.ascii_letters + string.digits))
        )),
    ]

    """
    Minimum character number of different character types required in check_character_set validation
    """
    required_min_char_types: int = 3

    def check_length(self, pw: str, user):
        """
        Raises PasswordPolicyError if a password is not at least min_length characters long.
        :param pw: password to check
        :param user: user entity
        """
        if len(pw) < self.min_length:
            raise PasswordPolicyError(_(
                'Password must be at least {min_length} characters long',
                min_length=self.min_length
            ))

    def check_character_set(self, pw: str, user):
        """
        Raises PasswordPolicyError if a password does not contain at least one character from
        at least `required_at_least_char_types` of the alphabets in `required_char_sets`.
        :param pw: password to check
        :param user: user entity
        """

        missing = []
        for name, alphabet in self.required_char_types:
            if not set(pw) & set(alphabet):
                missing.append(name)

        if len(missing) > len(self.required_char_types) - self.required_min_char_types:
            if len(self.required_char_types) == 1:
                message = _('Password must include a {type}', type=self.required_char_types[0].name)
            else:
                first_part = ', '.join(str(t.name) for t in self.required_char_types[:-1])
                message = _(
                    'Password must include at least {required} of {first} and/or {last}',
                    required=self.required_min_char_types,
                    first=first_part,
                    last=self.required_char_types[-1].name
                )

            raise PasswordPolicyError(message)

    def check_does_not_contain_username(self, pw: str, user):
        """
        Raises PasswordPolicyError if the password contains the username. This is case insensitive.
        :param pw: password to check
        :param user: user entity
        """
        user_cls = user.__class__
        username_key = get_username_key(user_cls)
        username = getattr(user, username_key)
        username_col = getattr(user_cls, username_key)
        if isinstance(username_col.type, EmailType):
            username = user.username.split('@')[0]

        if username.casefold() in pw.casefold():
            raise PasswordPolicyError(_('Password may not contain username'))

    def password_checks(self) -> typing.List[PasswordCheckFunction]:
        return [
            self.check_length,
            self.check_character_set,
            self.check_does_not_contain_username,
        ]

    @classmethod
    def generate_validator(cls, check: PasswordCheckFunction) -> typing.Callable:
        def validator(form: wtforms.Form, field: wtforms.Field):
            try:
                check(field.data, form.user)
            except PasswordPolicyError as e:
                raise wtforms.ValidationError(str(e))
        return validator

    @classmethod
    def form_validators(cls):
        policy = cls()
        return [cls.generate_validator(c) for c in policy.password_checks()]


class DefaultPasswordPolicy(PasswordPolicy):
    """
    A bare-bones, very permissive policy to use as a default if none is set on initialization.
    """
    def password_checks(self):
        return [self.check_length]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import keg_auth
import datetime as dt

import configparser


# -- Project information -----------------------------------------------------

project = 'Keg-Auth'
copyright = u"{year} Level 12".format(year=dt.datetime.utcnow().year)


cfg = configparser.SafeConfigParser()
cfg.read('../../setup.cfg')

tag = cfg.get('egg_info', 'tag_build')
master_doc = "index"

html_context = {
    'prerelease': bool(tag),  # True if tag is not the empty string
}

# The full version, including alpha/beta/rc tags.
release = keg_auth.__version__ + tag


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

html_theme_options = {
    'github_user': 'level12',
    'github_repo': 'keg-auth',
    'github_banner': False,
    'github_button': True,
    'codecov_button': True,
    'extra_nav_links': {
        'Level 12': 'https://www.level12.io',
        'File an Issue': 'https://github.com/level12/keg-auth/issues/new',
    },
    'show_powered_by': True
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



from unittest import mock

import flask
import flask_login
import freezegun
import arrow
from keg_auth.core import update_last_login
from keg_auth_ta.model import entities as ents


class TestUpdateLastLogin(object):

    @freezegun.freeze_time("2018-10-01 15:00:00")
    def test_it_will_update_last_login(self):
        u = ents.User.fake(email=u"test@bar.com", last_login_utc=None)
        update_last_login(flask.current_app, u)
        ents.db.session.remove()
        u = ents.User.get_by(email=u"test@bar.com")
        assert u.last_login_utc == arrow.utcnow()


class TestSessionClear:
    def test_session_cleared_after_signing_out(self):
        with flask.current_app.test_request_context():
            flask.session['foo'] = 'bar'
            assert 'foo' in list(flask.session)
            flask_login.logout_user()
            assert len(list(flask.session)) == 0

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_LOGOUT_CLEAR_SESSION': False})
    def test_session_not_cleared_after_signing_out(self):
        with flask.current_app.test_request_context():
            flask.session['foo'] = 'bar'
            assert 'foo' in list(flask.session)
            flask_login.logout_user()
            assert len(list(flask.session)) == 1



import flask
import flask_login
from markupsafe import Markup
import webgrid
from webgrid import filters

from keg_auth.extensions import lazy_gettext as _
from keg_auth.libs.templates import link_to, render_jinja
from keg_auth.model.utils import has_permissions
from flask_wtf.csrf import generate_csrf


class ActionColumn(webgrid.Column):
    """Places various action buttons in a Column.

    :param edit_permission_for: is a function that takes a row and returns the permission
                                required to open the edit endpoint for that row.
    :param delete_permission_for: is like `edit_permission_for`, but for the delete endpoint.
    :param view_permission_for: is like `edit_permission_for`, but for the view endpoint.
    :param view_link_class_for: is a function that takes a row and returns the HTML class to
                                place on the view link.
    :param edit_link_class_for: is a function that takes a row and returns the HTML class to
                                place on the edit link.
    :param delete_link_class_for: is a function that takes a row and returns the HTML class to
                                    place on the delete link.
    """
    default_view_link_class = 'view-link'
    default_edit_link_class = 'edit-link'
    default_delete_link_class = 'delete-link confirm-delete'

    def __init__(self,
                 label,
                 key=None,
                 filter=None,
                 can_sort=False,
                 render_in=('html',),
                 has_subtotal=False,
                 edit_endpoint=None,
                 delete_endpoint=None,
                 view_endpoint=None,
                 edit_permission_for=lambda row: None,
                 delete_permission_for=lambda row: None,
                 view_permission_for=lambda row: None,
                 view_link_class_for=None,
                 edit_link_class_for=None,
                 delete_link_class_for=None,
                 **kwargs):
        def link_class_for_fun(default_link_class):
            def link_class_for(row):
                return default_link_class

            return link_class_for

        if view_link_class_for is None:
            view_link_class_for = link_class_for_fun(self.default_view_link_class)
        if edit_link_class_for is None:
            edit_link_class_for = link_class_for_fun(self.default_edit_link_class)
        if delete_link_class_for is None:
            delete_link_class_for = link_class_for_fun(self.default_delete_link_class)

        self.edit_endpoint = edit_endpoint
        self.delete_endpoint = delete_endpoint
        self.view_endpoint = view_endpoint
        self.edit_permission_for = edit_permission_for
        self.delete_permission_for = delete_permission_for
        self.view_permission_for = view_permission_for
        self.view_link_class_for = view_link_class_for
        self.edit_link_class_for = edit_link_class_for
        self.delete_link_class_for = delete_link_class_for

        super(ActionColumn, self).__init__(label, key=key, filter=filter, can_sort=can_sort,
                                           render_in=render_in, has_subtotal=has_subtotal, **kwargs)

    def extract_and_format_data(self, record):
        view_perm = self.view_permission_for(record)
        edit_perm = self.edit_permission_for(record)
        delete_perm = self.delete_permission_for(record)
        can_edit = has_permissions(edit_perm, flask_login.current_user)
        can_delete = has_permissions(delete_perm, flask_login.current_user)
        can_view = (
            (self.edit_endpoint != self.view_endpoint or not can_edit)
            and has_permissions(view_perm, flask_login.current_user)
        )

        view_link_class = self.view_link_class_for(record)
        edit_link_class = self.edit_link_class_for(record)
        delete_link_class = self.delete_link_class_for(record)
        data = self.extract_data(record)
        return self.format_data(data, can_edit, can_delete, can_view,
                                view_link_class, edit_link_class, delete_link_class)

    def format_data(self, value, show_edit, show_delete, show_view,
                    view_link_class, edit_link_class, delete_link_class):
        result = Markup()
        if self.edit_endpoint and show_edit:
            result += link_to(
                Markup('&nbsp;'),
                flask.url_for(self.edit_endpoint, objid=value, session_key=self.grid.session_key),
                **{
                    'aria-label': _('Edit'),
                    'class': edit_link_class,
                    'title': _('Edit')
                }
            )
        if self.delete_endpoint and show_delete:
            result += link_to(
                Markup('&nbsp;'),
                flask.url_for(self.delete_endpoint, objid=value, session_key=self.grid.session_key),
                **{
                    'aria-label': _('Delete'),
                    'class': delete_link_class,
                    'title': _('Delete')
                }
            )
        if self.view_endpoint and show_view:
            result += link_to(
                Markup('&nbsp;'),
                flask.url_for(self.view_endpoint, objid=value, session_key=self.grid.session_key),
                **{
                    'aria-label': _('View'),
                    'class': view_link_class,
                    'title': _('View')
                }
            )
        return result


def make_user_grid(edit_endpoint, edit_permission, delete_endpoint, delete_permission,
                   grid_cls=None, resend_verification_endpoint=None):
    """Factory method to create a User grid class for CRUD."""
    user_cls = flask.current_app.auth_manager.entity_registry.user_cls
    grid_cls = grid_cls or flask.current_app.auth_manager.grid_cls
    action_column_cls = getattr(grid_cls, 'action_column_cls', ActionColumn)

    class ResendVerificationColumn(webgrid.Column):

        def __init__(self,
                     label,
                     url,
                     key=None,
                     render_in=('html',),
                     **kwargs):
            self.url = url
            super(ResendVerificationColumn, self).__init__(
                label, key=key, filter=None, can_sort=False,
                render_in=render_in, has_subtotal=False)

        def extract_and_format_data(self, record):
            return self.format_data(record)

        def format_data(self, data):
            result = Markup()
            if data.is_verified is False:
                result += render_jinja(
                    '<form action="{{ url }}" method="post">'
                    '<input type="hidden" name="csrf_token" value="{{ csrf_token }}" />'
                    '<input type="hidden" name="user_id" value="{{ user_id }}" />'
                    '<input type="submit" class"btn btn-primary" value="{{ submit_label }}" />'
                    '</form>',
                    url=flask.url_for(self.url),
                    csrf_token=generate_csrf(),
                    user_id=data.id,
                    submit_label=self.label,
                )
            return result

    class User(grid_cls):
        action_column_cls(
            '',
            user_cls.id,
            edit_endpoint=edit_endpoint,
            delete_endpoint=delete_endpoint,
            edit_permission_for=lambda _: edit_permission,
            delete_permission_for=lambda _: delete_permission
        )
        webgrid.Column(_('User ID'), user_cls.username, filters.TextFilter)
        if flask.current_app.auth_manager.mail_manager and hasattr(user_cls, 'is_verified'):
            webgrid.YesNoColumn(_('Verified'), user_cls.is_verified, filters.YesNoFilter)
        webgrid.YesNoColumn(_('Superuser'), user_cls.is_superuser, filters.YesNoFilter)
        if (
            flask.current_app.auth_manager.mail_manager
            and hasattr(user_cls, 'is_verified')
            and resend_verification_endpoint is not None
            and flask.current_app.config['KEGAUTH_EMAIL_OPS_ENABLED']
        ):
            ResendVerificationColumn(_('Resend Verification'), resend_verification_endpoint)

        def query_prep(self, query, has_sort, has_filters):
            if not has_sort:
                query = query.order_by(user_cls.username)
            return query
    return User


def make_group_grid(edit_endpoint, edit_permission, delete_endpoint, delete_permission,
                    grid_cls=None):
    """Factory method to create a Group grid class for CRUD."""
    group_cls = flask.current_app.auth_manager.entity_registry.group_cls
    grid_cls = grid_cls or flask.current_app.auth_manager.grid_cls
    action_column_cls = getattr(grid_cls, 'action_column_cls', ActionColumn)

    class Group(grid_cls):
        action_column_cls(
            '',
            group_cls.id,
            edit_endpoint=edit_endpoint,
            delete_endpoint=delete_endpoint,
            edit_permission_for=lambda _: edit_permission,
            delete_permission_for=lambda _: delete_permission
        )
        webgrid.Column(_('Name'), group_cls.name, filters.TextFilter)

        def query_prep(self, query, has_sort, has_filters):
            if not has_sort:
                query = query.order_by(group_cls.name)
            return query
    return Group


def make_bundle_grid(edit_endpoint, edit_permission, delete_endpoint, delete_permission,
                     grid_cls=None):
    """Factory method to create a Bundle grid class for CRUD."""
    bundle_cls = flask.current_app.auth_manager.entity_registry.bundle_cls
    grid_cls = grid_cls or flask.current_app.auth_manager.grid_cls
    action_column_cls = getattr(grid_cls, 'action_column_cls', ActionColumn)

    class Bundle(grid_cls):
        action_column_cls(
            '',
            bundle_cls.id,
            edit_endpoint=edit_endpoint,
            delete_endpoint=delete_endpoint,
            edit_permission_for=lambda _: edit_permission,
            delete_permission_for=lambda _: delete_permission
        )
        webgrid.Column(_('Name'), bundle_cls.name, filters.TextFilter)

        def query_prep(self, query, has_sort, has_filters):
            if not has_sort:
                query = query.order_by(bundle_cls.name)
            return query
    return Bundle


def make_permission_grid(grid_cls=None):
    """Factory method to create a Permission grid class."""
    permission_cls = flask.current_app.auth_manager.entity_registry.permission_cls
    grid_cls = grid_cls or flask.current_app.auth_manager.grid_cls

    class Permission(grid_cls):
        webgrid.Column(_('Name'), permission_cls.token, filters.TextFilter)
        webgrid.Column(_('Description'), permission_cls.description, filters.TextFilter)

        def query_prep(self, query, has_sort, has_filters):
            if not has_sort:
                query = query.order_by(permission_cls.token)
            return query
    return Permission



import logging

import flask
import flask_login
import keg.web

from keg_auth import make_blueprint, requires_permissions, requires_user, has_any, has_all
from keg_auth.forms import user_form
from keg_auth.views import User as UserBase
from keg_auth_ta.extensions import csrf, auth_manager

log = logging.getLogger(__name__)


@requires_permissions('permission1')
class ProtectedBlueprint(flask.Blueprint):
    def on_authentication_failure(self):
        flask.abort(405)


@requires_permissions(lambda user: hasattr(user, 'blueprint_pass'))
class CallableProtectedBlueprint(flask.Blueprint):
    pass


@requires_user(http_methods_excluded=['OPTIONS'])
class ProtectedBlueprint2(flask.Blueprint):
    pass


class User(UserBase):
    # need to make this a static method so it isn't bound on the view instance
    form_cls = staticmethod(user_form)

    def pre_method(self):
        self.assign('a_template_variable', 'value present')

    def create_form(self, obj):
        form_cls = self.form_cls(flask.current_app.config,
                                 allow_superuser=flask_login.current_user.is_superuser,
                                 endpoint=self.endpoint_for_action('edit'))
        return form_cls(obj=obj)


public_bp = flask.Blueprint('public', __name__)
private_bp = flask.Blueprint('private', __name__)
protected_bp = ProtectedBlueprint('protected', __name__)
protected_bp2 = ProtectedBlueprint2('protected2', __name__)
callable_protected_bp = CallableProtectedBlueprint('callable_protected', __name__)
auth_bp = make_blueprint(__name__, auth_manager, user_crud_cls=User)

blueprints = public_bp, private_bp, protected_bp, protected_bp2, callable_protected_bp, auth_bp

# Exempt from CSRF or we have problems with Secret2.post.
csrf.exempt(private_bp)


class Home(keg.web.BaseView):
    blueprint = public_bp
    url = '/'
    template_name = 'home.html'
    nav_group = 'auth'
    auto_assign = ['nav_group']

    def get(self):
        pass


@private_bp.route('/secret1')
@requires_user
def secret1():
    return 'secret1'


@requires_user()
class Secret1Class(keg.web.BaseView):
    """ Show class decorator usage of requires_user, and also that the decorator works with or
        without ()
    """
    blueprint = private_bp

    def get(self):
        return 'secret1-class'

    def on_authentication_failure(self):
        flask.abort(405)


@requires_user
class Secret1FlaskClass(flask.views.MethodView):
    def get(self):
        return 'secret1-flask-class'

    def on_authentication_failure(self):
        flask.abort(405)


private_bp.add_url_rule(
    '/secret1-flask-class', view_func=Secret1FlaskClass.as_view('secret1-flask-class'))


class Secret2(keg.web.BaseView):
    blueprint = private_bp

    @requires_permissions(has_any('permission1', 'permission2'))
    def get(self):
        return 'secret2'

    @requires_permissions(has_all('permission1', 'permission2'))
    def post(self):
        return 'secret2 post'

    @requires_permissions('permission1')
    def put(self):
        return 'secret2 put'

    @requires_permissions('permission1')
    def patch(self):
        return 'secret2 patch'

    @requires_permissions('permission1')
    def delete(self):
        return 'secret2 delete'

    @requires_permissions('permission1')
    def options(self):
        return 'secret2 options'


@requires_permissions(has_all('permission1', 'permission2'))
class Secret3(keg.web.BaseView):
    blueprint = private_bp

    def get(self):
        return 'secret3'


class Secret3Sub(Secret3):
    def get(self):
        return 'secret3-sub'


@requires_permissions(has_all('permission1', 'permission2'))
class SecretFlask(flask.views.MethodView):
    def get(self):
        return 'secret-flask'


private_bp.add_url_rule(
    '/secret-flask', view_func=SecretFlask.as_view('secret-flask'))


class SecretFlaskSub(SecretFlask):
    def get(self):
        return 'secret-flask-sub'


private_bp.add_url_rule(
    '/secret-flask-sub', view_func=SecretFlaskSub.as_view('secret-flask-sub'))


@requires_permissions('permission1')
class Secret4(keg.web.BaseView):
    blueprint = private_bp

    @requires_permissions('permission2')
    def get(self):
        return 'secret4'

    def on_authorization_failure(self):
        flask.abort(405)


@requires_permissions('permission1')
class SecretFlask4(flask.views.MethodView):

    @requires_permissions('permission2')
    def get(self):
        return 'secret-flask4'

    def on_authorization_failure(self):
        flask.abort(405)


private_bp.add_url_rule(
    '/secret-flask4', view_func=SecretFlask4.as_view('secret-flask4'))


@private_bp.route('/secret-nested')
@requires_permissions(has_any(has_all('permission1', 'permission2'), 'permission3'))
def secret_nested():
    return 'secret_nested'


@private_bp.route('/secret-callable')
@requires_permissions(lambda user: user.email == 'foo@bar.baz')
def secret_callable():
    return 'secret_callable'


@private_bp.route('/secret-nested-callable')
@requires_permissions(has_any('permission1',
                              lambda user: user.email == 'foo@bar.baz'))
def secret_nested_callable():
    return 'secret_nested_callable'


@requires_permissions(lambda user: hasattr(user, 'class_pass'))
class CallableProtectedClass(keg.web.BaseView):
    blueprint = callable_protected_bp

    def get(self):
        return 'callable-protected-class'


class SecretNavURLOnClass(keg.web.BaseView):
    blueprint = private_bp

    @private_bp.route('/secret-route-on-class')
    @requires_permissions('permission1')
    def someroute(self):
        return 'secret-route-on-class'


class ProtectedClass(keg.web.BaseView):
    blueprint = protected_bp

    def get(self):
        return 'protected-class'


class ProtectedClass2(keg.web.BaseView):
    blueprint = protected_bp2

    def get(self):
        return 'protected-class2'


@protected_bp.route('/protected-method')
def protected_method():
    return 'protected-method'


@private_bp.route('/jwt-required')
@requires_user
def jwt_required():
    return 'jwt-required'


@private_bp.route('/custom-auth-failure')
@requires_user(on_authentication_failure=lambda: flask.abort(400))
def custom_auth_failure():
    return 'custom-auth-failure'


@private_bp.route('/custom-perm-failure')
@requires_permissions('permission1', on_authorization_failure=lambda: flask.abort(400))
def custom_perm_failure():
    return 'custom-perm-failure'



import os.path as osp

from setuptools import setup, find_packages

cdir = osp.abspath(osp.dirname(__file__))
README = open(osp.join(cdir, 'readme.rst')).read()
CHANGELOG = open(osp.join(cdir, 'changelog.rst')).read()

version_fpath = osp.join(cdir, 'keg_auth', 'version.py')
version_globals = {}
with open(version_fpath) as fo:
    exec(fo.read(), version_globals)

setup(
    name='Keg-Auth',
    version=version_globals['VERSION'],
    description='Authentication plugin for Keg',
    long_description='\n\n'.join((README, CHANGELOG)),
    long_description_content_type='text/x-rst',
    author='Randy Syring',
    author_email='randy.syring@level12.io',
    url='https://github.com/level12/keg-auth',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
    packages=find_packages(exclude=[]),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'authlib',
        'bcrypt',
        'markdown-it-py',

        'Flask-Login>0.4.1',
        'Keg>=0.10.2',
        'KegElements>=0.8.0',
        'inflect',
        'passlib',
        'shortuuid',
        'webgrid>=0.4.13',
    ],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'test': [
            'email_validator',
            'flake8',
            'flask-bootstrap',
            'flask-jwt-extended>=4.0.0',
            'flask-mail',
            'flask-webtest',
            'freezegun',
            'mock',
            'psycopg2-binary',
            'python-ldap==3.4.2',
            'pyquery',
            'pytest',
            'pytest-cov',
            'requests',
            'tox',
            'xlsxwriter',
        ],
        'i18n': [
            'morphi',
            'webgrid[i18n]'
        ],
        'jwt': [
            'flask-jwt-extended>=4.0.0',
        ],
        'ldap': [
            'python-ldap',
        ],
        'mail': [
            'Flask-Mail',
        ],
        'oauth': [
            'requests',
        ],
    }
)
